using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;

namespace EdgeDetectionMethods
{
    public class ConvolutionMatrix
    {
        private int tLeft = 0, tCenter = 0, tRigth = 0, mLeft = 0, mCenter = 0, mRight = 0, bLeft = 0, bCenter = 0, bRight = 0;
        private int iFactor = 1, iOffset = 0;

        public void SetMatrixValues(int pos00, int pos01, int pos02, int pos10, int pos11, int pos12, int pos20, int pos21, int pos22)
        {
            tLeft = pos00;
            tCenter = pos01;
            tRigth = pos02;
            mLeft = pos10;
            mCenter = pos11;
            mRight = pos12;
            bLeft = pos20;
            bCenter = pos21;
            bRight = pos22;
        }

        public void setCenterOnly(int iVal)
        {
            mCenter = iVal;
        }

        public void setFactor(int iVal)
        {
            iFactor = iVal;
        }

        public void setOffset(int iVal)
        {
            iOffset = iVal;
        }

        public int getOffset()
        {
            return iOffset;
        }

        public int getFactor()
        {
            return iFactor;
        }

        public int getValue(int pos)
        {
            switch (pos)
            {
                case 1: return tLeft;
                case 2: return tCenter;
                case 3: return tRigth;
                case 4: return mLeft;
                case 5: return mCenter;
                case 6: return mRight;
                case 7: return bLeft;
                case 8: return bCenter;
                case 9: return bRight;
            }
            return 0;
        }
    }


    public class EdgeDetectionMethods
    {

        public static bool Convolution3x3(Bitmap bitmapOriginal, ConvolutionMatrix matrix)
        {
            //string strMatrix = matrix.getValue(1).ToString() + matrix.getValue(2).ToString() + matrix.getValue(3).ToString() + "\n" + matrix.getValue(4).ToString() + matrix.getValue(5).ToString() + matrix.getValue(6).ToString() + "\n" + matrix.getValue(7).ToString() + matrix.getValue(8).ToString() + matrix.getValue(9).ToString();
            //System.Windows.Forms.MessageBox.Show(strMatrix);
            


            if (0 == matrix.getFactor()) 
                return false;

            Bitmap bitmapSource = (Bitmap)bitmapOriginal.Clone();

            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bitmapDataSource = bitmapSource.LockBits(new Rectangle(0, 0, bitmapSource.Width, bitmapSource.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            int iStride = bitmapDataOriginal.Stride;
            int iDoubleiStride = iStride * 2;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;
            System.IntPtr iPtrScanSource = bitmapDataSource.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                byte* ptrSource = (byte*)(void*)iPtrScanSource;

                int iOffset = iStride - bitmapOriginal.Width * 3;
                int iWidth = bitmapOriginal.Width - 2;
                int iHeight = bitmapOriginal.Height - 2;

                int iPixel;
                //string strVal = "";


                for (int y = 0; y < iHeight; ++y)
                {
                    for (int x = 0; x < iWidth; ++x)
                    {
                        iPixel = ((((ptrSource[2] * matrix.getValue(1)) + (ptrSource[5] * matrix.getValue(2)) + (ptrSource[8] * matrix.getValue(3)) +
                            (ptrSource[2 + iStride] * matrix.getValue(4)) + (ptrSource[5 + iStride] * matrix.getValue(5)) + (ptrSource[8 + iStride] * matrix.getValue(6)) +
                            (ptrSource[2 + iDoubleiStride] * matrix.getValue(7)) + (ptrSource[5 + iDoubleiStride] * matrix.getValue(8)) + (ptrSource[8 + iDoubleiStride] * matrix.getValue(9))) / matrix.getFactor()) + matrix.getOffset());

                        if (iPixel < 0) 
                            iPixel = 0;
                        if (iPixel > 255) 
                            iPixel = 255;
                        //if (x < 25 & y < 25)
                        //strVal = "258=" + iPixel.ToString();

                        ptrOriginal[5 + iStride] = (byte)iPixel;

                        iPixel = ((((ptrSource[1] * matrix.getValue(1)) + (ptrSource[4] * matrix.getValue(2)) + (ptrSource[7] * matrix.getValue(3)) +
                            (ptrSource[1 + iStride] * matrix.getValue(4)) + (ptrSource[4 + iStride] * matrix.getValue(5)) + (ptrSource[7 + iStride] * matrix.getValue(6)) +
                            (ptrSource[1 + iDoubleiStride] * matrix.getValue(7)) + (ptrSource[4 + iDoubleiStride] * matrix.getValue(8)) + (ptrSource[7 + iDoubleiStride] * matrix.getValue(9))) / matrix.getFactor()) + matrix.getOffset());

                        if (iPixel < 0) 
                            iPixel = 0;
                        if (iPixel > 255) 
                            iPixel = 255;
                        //if (x < 25 & y < 25)
                        //strVal = strVal +"\n147="+ iPixel.ToString();

                        ptrOriginal[4 + iStride] = (byte)iPixel;
                        /*int a = ptrSource[0];
                        int b = ptrSource[3];
                        int c = ptrSource[6];
                        int o = ptrSource[0 + iStride];
                        int p = ptrSource[3 + iStride];
                        int q = ptrSource[6 + iStride];
                        int r = ptrSource[0 + iDoubleiStride];
                        int s = ptrSource[3 + iDoubleiStride];
                        int t = ptrSource[6 + iDoubleiStride];
                        int d = matrix.getValue(1);
                        int e = matrix.getValue(2);
                        int f = matrix.getValue(3);
                        int g = matrix.getValue(4);
                        int h = matrix.getValue(5);
                        int i = matrix.getValue(6);
                        int j = matrix.getValue(7);
                        int k = matrix.getValue(8);
                        int l = matrix.getValue(9);
                        int m = matrix.getOffset();
                        int n = matrix.getFactor();*/

                        //strVal = strVal + "\nInputTo036=iStride" + iStride.ToString() + " iDoubleiStride"+ iDoubleiStride.ToString();
                        iPixel = ((((ptrSource[0] * matrix.getValue(1)) + (ptrSource[3] * matrix.getValue(2)) + (ptrSource[6] * matrix.getValue(3)) +
                            (ptrSource[0 + iStride] * matrix.getValue(4)) + (ptrSource[3 + iStride] * matrix.getValue(5)) + (ptrSource[6 + iStride] * matrix.getValue(6)) +
                            (ptrSource[0 + iDoubleiStride] * matrix.getValue(7)) + (ptrSource[3 + iDoubleiStride] * matrix.getValue(8)) + (ptrSource[6 + iDoubleiStride] * matrix.getValue(9))) / matrix.getFactor()) + matrix.getOffset());

                        if (iPixel < 0) 
                            iPixel = 0;
                        if (iPixel > 255) 
                            iPixel = 255;
                        //if (x < 25 && y < 25)
                        //strVal = strVal +"\n036="+ iPixel.ToString();

                        //if(x < 25 & y < 25)
                        //System.Windows.Forms.MessageBox.Show(strVal);

                        //strVal = "";

                        ptrOriginal[3 + iStride] = (byte)iPixel;

                        ptrOriginal += 3;
                        ptrSource += 3;
                    }
                    ptrOriginal += iOffset;
                    ptrSource += iOffset;
                }
               
            }

            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            bitmapSource.UnlockBits(bitmapDataSource);

            return true;
        }

        public static bool GrayScale(Bitmap bitmapOriginal)
        {
            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            int stride = bitmapDataOriginal.Stride;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                int nOffset = stride - bitmapOriginal.Width * 3;
                byte red, green, blue;

                for (int y = 0; y < bitmapOriginal.Height; ++y)
                {
                    for (int x = 0; x < bitmapOriginal.Width; ++x)
                    {
                        blue = ptrOriginal[0];
                        green = ptrOriginal[1];
                        red = ptrOriginal[2];

                        ptrOriginal[0] = ptrOriginal[1] = ptrOriginal[2] = (byte)(.299 * red + .587 * green + .114 * blue);

                        ptrOriginal += 3;
                    }
                    ptrOriginal += nOffset;
                }
            }
            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            return true;
        }

        public static bool SepiaMicrosoft(Bitmap bitmapOriginal)
        {
            //needs work...output not matching what is at http://blogs.techrepublic.com.com/howdoi/?p=120

            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            int stride = bitmapDataOriginal.Stride;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                int nOffset = stride - bitmapOriginal.Width * 3;
                byte red, green, blue;

                for (int y = 0; y < bitmapOriginal.Height; ++y)
                {
                    for (int x = 0; x < bitmapOriginal.Width; ++x)
                    {
                        
                        blue = ptrOriginal[0];
                        green = ptrOriginal[1];
                        red = ptrOriginal[2];

                        double blueColor = (.272 * red + .534 * green + .131 * blue);
                        double greenColor = (.349 * red + .686 * green + .168 * blue);
                        double redColor = (.393 * red + .769 * green + .189 * blue);

                        ptrOriginal[0] = (byte)blueColor;
                        ptrOriginal[1] = (byte)greenColor;
                        ptrOriginal[2] = (byte)redColor;

                        if (blueColor > 255)
                            ptrOriginal[0] = 255;
                        if (greenColor > 255)
                            ptrOriginal[1] = 255;
                        if (redColor > 255)
                            ptrOriginal[2] = 255;

                        ptrOriginal += 3;
                    }
                    ptrOriginal += nOffset;
                }
            }
            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            return true;
        }

        public static bool blurImage(Bitmap bitmapOriginal, int Intensity)
        {
            ConvolutionMatrix matrix = new ConvolutionMatrix();
            matrix.SetMatrixValues(1, 2, 1, 2, Intensity, 2, 1, 2, 1);
            matrix.setFactor(Intensity + 12);
            return EdgeDetectionMethods.Convolution3x3(bitmapOriginal,matrix);
        }



        public static bool BasicEdgeDetect(Bitmap bitmapOriginal, byte nThreshold, string strType, int[] iMatrix)
        {
            ConvolutionMatrix matrix = new ConvolutionMatrix();
            Bitmap bitmapClone = (Bitmap)bitmapOriginal.Clone();

            switch (strType)
            {
                case "SOBEL":
                    matrix.SetMatrixValues(1, 0, -1, 2, 0, -2, 1, 0, -1);
                    //matrix.SetMatrixValues(1, 2, 1, 0, 0, 0, -1, -2, -1);
                    matrix.setOffset(0);
                    break;
                case "PREWITT":
                    matrix.SetMatrixValues(-1, 0, 1, -1, 0, 1, -1, 0, 1);
                    //matrix.SetMatrixValues(1, 1, 1, 0, 0, 0, -1, -1, -1);
                    matrix.setOffset(0);
                    break;
                case "KIRSH":
                    matrix.SetMatrixValues(5, -3, -3, 5, 0, -3, 5, -3, -3);
                    //matrix.SetMatrixValues(5, 5, 5, -3, -3, -3, -3, -3, -3);
                    matrix.setOffset(0);
                    break;
                case "CUSTOM":
                    matrix.SetMatrixValues(iMatrix[0], iMatrix[1], iMatrix[2], iMatrix[3], iMatrix[4], iMatrix[5], iMatrix[6], iMatrix[7], iMatrix[8]);
                    matrix.setOffset(iMatrix[9]);
                    matrix.setFactor(iMatrix[10]);
                    break;
            }

            EdgeDetectionMethods.Convolution3x3(bitmapOriginal, matrix);

            switch (strType)
            {
                case "SOBEL":
                    matrix.SetMatrixValues(1, 2, 1, 0, 0, 0, -1, -2, -1);
                    matrix.setOffset(0);
                    break;
                case "PREWITT":
                    matrix.SetMatrixValues(1, 1, 1, 0, 0, 0, -1, -1, -1);
                    matrix.setOffset(0);
                    break;
                case "KIRSH":
                    matrix.SetMatrixValues(-3, -3, -3, -3, 0, -3, 5, 5, 5);
                    matrix.setOffset(0);
                    break;
                case "CUSTOM":
                    matrix.SetMatrixValues(iMatrix[8], iMatrix[5], iMatrix[2], iMatrix[7], iMatrix[4], iMatrix[1], iMatrix[6], iMatrix[3], iMatrix[0]);
                    matrix.setOffset(iMatrix[9]);
                    matrix.setFactor(iMatrix[10]);
                    break;

            }

            EdgeDetectionMethods.Convolution3x3(bitmapClone, matrix);

            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bitmapDataSource = bitmapClone.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            int iStride = bitmapDataOriginal.Stride;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;
            System.IntPtr iPtrScanSource = bitmapDataSource.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                byte* ptrSource = (byte*)(void*)iPtrScanSource;
                int nOffset = iStride - bitmapOriginal.Width * 3;
                int nWidth = bitmapOriginal.Width * 3;
                int nPixel = 0;

                for (int y = 0; y < bitmapOriginal.Height; ++y)
                {
                    for (int x = 0; x < nWidth; ++x)
                    {
                        nPixel = (int)Math.Sqrt((ptrOriginal[0] * ptrOriginal[0]) + (ptrSource[0] * ptrSource[0]));
                        if (nPixel < nThreshold) 
                            nPixel = nThreshold;
                        if (nPixel > 255) 
                            nPixel = 255;
                        ptrOriginal[0] = (byte)nPixel;
                        ++ptrOriginal;
                        ++ptrSource;
                    }
                    ptrOriginal += nOffset;
                    ptrSource += nOffset;
                }
            }

            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            bitmapClone.UnlockBits(bitmapDataSource);

            return true;
        }

        public static bool HorizontalEdgeDetection(Bitmap bitmapOriginal)
        {
            Bitmap bitmapClone = (Bitmap)bitmapOriginal.Clone();

            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bitmapDataSource = bitmapClone.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            int iStride = bitmapDataOriginal.Stride;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;
            System.IntPtr iPtrScanSource = bitmapDataSource.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                byte* ptrSource = (byte*)(void*)iPtrScanSource;

                int nOffset = iStride - bitmapOriginal.Width * 3;
                int nWidth = bitmapOriginal.Width * 3;

                int nPixel = 0;

                ptrOriginal += iStride;
                ptrSource += iStride;

                for (int y = 1; y < bitmapOriginal.Height - 1; ++y)
                {
                    ptrOriginal += 9;
                    ptrSource += 9;

                    for (int x = 9; x < nWidth - 9; ++x)
                    {
                        nPixel = ((ptrSource + iStride - 9)[0] +
                            (ptrSource + iStride - 6)[0] +
                            (ptrSource + iStride - 3)[0] +
                            (ptrSource + iStride)[0] +
                            (ptrSource + iStride + 3)[0] +
                            (ptrSource + iStride + 6)[0] +
                            (ptrSource + iStride + 9)[0] -
                            (ptrSource - iStride - 9)[0] -
                            (ptrSource - iStride - 6)[0] -
                            (ptrSource - iStride - 3)[0] -
                            (ptrSource - iStride)[0] -
                            (ptrSource - iStride + 3)[0] -
                            (ptrSource - iStride + 6)[0] -
                            (ptrSource - iStride + 9)[0]);

                        if (nPixel < 0) nPixel = 0;
                        if (nPixel > 255) nPixel = 255;

                        (ptrOriginal + iStride)[0] = (byte)nPixel;

                        ++ptrOriginal;
                        ++ptrSource;
                    }

                    ptrOriginal += 9 + nOffset;
                    ptrSource += 9 + nOffset;
                }
            }

            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            bitmapClone.UnlockBits(bitmapDataSource);

            return true;
        }

        public static bool VerticalEdgeDetection(Bitmap bitmapOriginal)
        {
            Bitmap bitmapClone = (Bitmap)bitmapOriginal.Clone();

            BitmapData bitmapDataOriginal = bitmapOriginal.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bitmapDataSource = bitmapClone.LockBits(new Rectangle(0, 0, bitmapOriginal.Width, bitmapOriginal.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            int iStride = bitmapDataOriginal.Stride;
            System.IntPtr iPtrScanOriginal = bitmapDataOriginal.Scan0;
            System.IntPtr iPtrScanSource = bitmapDataSource.Scan0;

            unsafe
            {
                byte* ptrOriginal = (byte*)(void*)iPtrScanOriginal;
                byte* ptrSource = (byte*)(void*)iPtrScanSource;

                int nOffset = iStride - bitmapOriginal.Width * 3;
                int nWidth = bitmapOriginal.Width * 3;

                int nPixel = 0;

                int nStride2 = iStride * 2;
                int nStride3 = iStride * 3;

                ptrOriginal += nStride3;
                ptrSource += nStride3;

                for (int y = 3; y < bitmapOriginal.Height - 3; ++y)
                {
                    ptrOriginal += 3;
                    ptrSource += 3;

                    for (int x = 3; x < nWidth - 3; ++x)
                    {
                        nPixel = ((ptrSource + nStride3 + 3)[0] +
                            (ptrSource + nStride2 + 3)[0] +
                            (ptrSource + iStride + 3)[0] +
                            (ptrSource + 3)[0] +
                            (ptrSource - iStride + 3)[0] +
                            (ptrSource - nStride2 + 3)[0] +
                            (ptrSource - nStride3 + 3)[0] -
                            (ptrSource + nStride3 - 3)[0] -
                            (ptrSource + nStride2 - 3)[0] -
                            (ptrSource + iStride - 3)[0] -
                            (ptrSource - 3)[0] -
                            (ptrSource - iStride - 3)[0] -
                            (ptrSource - nStride2 - 3)[0] -
                            (ptrSource - nStride3 - 3)[0]);

                        if (nPixel < 0) nPixel = 0;
                        if (nPixel > 255) nPixel = 255;

                        ptrOriginal[0] = (byte)nPixel;

                        ++ptrOriginal;
                        ++ptrSource;
                    }

                    ptrOriginal += 3 + nOffset;
                    ptrSource += 3 + nOffset;
                }
            }

            bitmapOriginal.UnlockBits(bitmapDataOriginal);
            bitmapClone.UnlockBits(bitmapDataSource);

            return true;
        }

    }
}
